import pandas as pd
import numpy as np
import scipy
from .utils.util import _rga, _delta_function, _num, _den 
import matplotlib.pyplot as plt


def rge(xtrain, xtest, ytrain, ytest, model):
    """
     ### RANK GRADUATION EXPLAINABILITY (RGE) MEASURE ###
     Function for the RGE measure computation
     Inputs: xtrain, xtest, ytrain, ytest -> the train and test data selected for training and testing the model; 
             model -> a classification or regression model. For example RandomForestClassifier() in ensemble module of sklearn.
    """    
    xtrain = pd.DataFrame(xtrain).reset_index(drop=True)
    xtest = pd.DataFrame(xtest).reset_index(drop=True)
    ytrain = pd.DataFrame(ytrain).reset_index(drop=True)
    ytest = pd.DataFrame(ytest).reset_index(drop=True)
    rge_list = []
    model_full = model.fit(xtrain, ytrain)
    yhat = model_full.predict(xtest)
    for i in xtrain.columns:
         xtrain_rm = xtrain.drop(i, axis=1)
         xtest_rm = xtest.drop(i, axis=1)
         model_rm = model.fit(xtrain_rm, ytrain)
         yhat_rm = model_rm.predict(xtest_rm)
         rge_list.append(1-(_rga(yhat, yhat_rm)))
    rge_df = pd.DataFrame(rge_list, index=xtest.columns, columns=["RGE"]).sort_values(by="RGE", ascending=False)
    plt.figure(figsize=(10, 6))
    plt.barh(rge_df.index, rge_df["RGE"])
    plt.xlabel("RGE (Feature Importance)")
    plt.ylabel("Feature")
    plt.title("RGE")
    plt.show()
    return rge_df


def rge_num(yhat, yhat_xk):
    yhat = pd.DataFrame(yhat).reset_index(drop=True)
    yhat_xk = pd.DataFrame(yhat_xk).reset_index(drop=True)
    df = pd.concat([yhat,yhat_xk], axis=1)
    df.columns = ["yhat", "yhat_xk"]
    ryhat_xk = yhat_xk.rank(method="min")
    df["ryhat_xk"] = ryhat_xk
    support = df.groupby('ryhat_xk')['yhat'].mean().reset_index(name='support')
    rord = list(range(len(yhat)))
    for jj in range(len(rord)):
        for ii in range(len(support)):
                if df["ryhat_xk"][jj]== support['ryhat_xk'][ii]:
                    rord[jj] = support['support'][ii]
    vals = [[i, values] for i, values in enumerate(df["yhat_xk"])]
    ranks = [x[0] for x in sorted(vals, key= lambda item: item[1])]
    ystar = [rord[i] for i in ranks]
    I = list(range(len(ystar)))
    conc = 2*sum([I[i]*ystar[i] for i in range(len(I))])
    dec= 2*sum([sorted(df["yhat"], reverse=True)[i]*I[i] for i in range(len(I))]) 
    inc = 2*sum([sorted(df["yhat"])[i]*I[i] for i in range(len(I))]) 
    RGE_num=(conc-dec)
    return RGE_num 

def rge_den(yhat):
    yhat = pd.DataFrame(yhat)
    yhat.columns = ["yhat"]
    I = list(range(len(yhat)))
    dec= 2*sum([yhat.sort_values(by="yhat", ascending=False, ignore_index=True).iloc[i,0]*I[i] for i in range(len(I))]) 
    inc = 2*sum([yhat.sort_values(by="yhat", ignore_index=True).iloc[i,0]*I[i] for i in range(len(I))]) 
    RGE_den=(inc-dec)
    return RGE_den

def delta_function(data):
    return rge_den(data.iloc[:,0])-rge_num(data.iloc[:,0], data.iloc[:,1])

def rge_statistic_test(yhat, yhat_xk):
     """
     RGE based test for comparing the ordering of the ranks related to the full model with that of the
     reduced model without the predictor of interest
     """
     yhat = pd.DataFrame(yhat).reset_index(drop=True)
     yhat_xk = pd.DataFrame(yhat_xk).reset_index(drop=True)
     jk_mat = pd.concat([yhat, yhat_xk], axis=1, keys=["yhat", "yhat_xk"])
     n = len(jk_mat)
     index = np.arange(n)
     jk_results = []
     for i in range(n):
          jk_sample = jk_mat.iloc[[x for x in index if x != i],:]
          jk_sample.reset_index(drop=True, inplace=True)
          jk_statistic = _delta_function(jk_sample)
          jk_results.append(jk_statistic)
     se = np.sqrt(((n-1)/n)*(sum([(x-np.mean(jk_results))**2 for x in jk_results])))
     z = (_den(yhat)-_num(yhat,yhat_xk))/se
     p_value = 2*scipy.stats.norm.cdf(-abs(z))
     return p_value


import pandas as pd
import numpy as np
import scipy
from .utils.util import _rga, _delta_function, _num, _den

def rgf(xtrain, xtest, ytrain, ytest, protectedvariables, model):
    """
     ### RANK GRADUATION FAIRNESS (RGF) MEASURE ###
     Function for the RGF measure computation
     Inputs: xtrain -> A dataframe including xtrain data used for traing the model; 
             xtest -> A dataframe including xtest data used for testing the model;
             ytrain -> Actual values of the target variable for traing the model; 
             ytest -> Actual values of the target variable for testing the model;
             protectedvariables -> A list of variables should be considered as the protected variables; 
             model -> Model could be a regressor such as MLPRegressor() from sklearn.neural_network
             or a classifier for example RandomForestClassifier() from sklearn.ensemble;
    """    
    if not isinstance(protectedvariables, list):
        raise ValueError("Protectedvariables input must be a list")
    xtrain = pd.DataFrame(xtrain).reset_index(drop=True)
    xtest = pd.DataFrame(xtest).reset_index(drop=True)
    ytrain = pd.DataFrame(ytrain).reset_index(drop=True)
    ytest = pd.DataFrame(ytest).reset_index(drop=True)
    for i in range(len(list(protectedvariables))):
        if protectedvariables[i] not in xtrain:
            raise ValueError(f"{protectedvariables[i]} is not in the variables")
    rgf_list = []
    model_full = model.fit(xtrain, ytrain)
    yhat = model_full.predict(xtest)
    for i in protectedvariables:
         xtrain_pr = xtrain.drop(i, axis=1)
         xtest_pr = xtest.drop(i, axis=1)
         model_pr = model.fit(xtrain_pr, ytrain)
         yhat_pr = model_pr.predict(xtest_pr)
         rgf_list.append(_rga(yhat, yhat_pr))
    return pd.DataFrame(rgf_list, index=protectedvariables, columns=["RGF"]).sort_values(by="RGF", ascending=False)


def rgf_statistic_test(yhat, yhat_rm):
        """
        RGF based test for comparing the ordering of the ranks related to the full model with that of the
        reduced model without the protected variable of interest
        """
        yhat = pd.DataFrame(yhat).reset_index(drop=True)
        yhat_rm = pd.DataFrame(yhat_rm).reset_index(drop=True)
        jk_mat = pd.concat([yhat, yhat_rm], axis=1, keys=["yhat", "yhat_rm"])
        n = len(jk_mat)
        index = np.arange(n)
        jk_results = []
        for i in range(n):
            jk_sample = jk_mat.iloc[[x for x in index if x != i],:]
            jk_sample.reset_index(drop=True, inplace=True)
            jk_statistic = _delta_function(jk_sample)
            jk_results.append(jk_statistic)
        se = np.sqrt(((n-1)/n)*(sum([(x-np.mean(jk_results))**2 for x in jk_results])))
        z = (_den(yhat)-_num(yhat, yhat_rm))/se
        p_value = 2*scipy.stats.norm.cdf(-abs(z))
        return p_value


import pandas as pd
import numpy as np
import scipy
from .utils.util import _rga, _delta_function, _num, _den


def rgp(xtrain, xtest, ytrain, ytest, index, model):
    """
     ### RANK GRADUATION PRIVACY (RGP) MEASURE ###
     Function for the RGP measure computation
     Inputs: xtrain -> A dataframe including xtrain data used for traing the model; 
             xtest -> A dataframe including xtest data used for testing the model;
             ytrain -> Actual values of the target variable for traing the model; 
             ytest -> Actual values of the target variable for testing the model;
             index -> A list of indices of the observations for which privacy is evaluated; 
             model -> Model could be a regressor such as MLPRegressor() from sklearn.neural_network
             or a classifier for example RandomForestClassifier() from sklearn.ensemble;
    """      
    if not isinstance(index, list):
        raise ValueError("Index input must be a list")
    xtrain = pd.DataFrame(xtrain).reset_index(drop=True)
    xtest = pd.DataFrame(xtest).reset_index(drop=True)
    ytrain = pd.DataFrame(ytrain).reset_index(drop=True)
    ytest = pd.DataFrame(ytest).reset_index(drop=True)
    rgp_list = []
    model_full = model.fit(xtrain, ytrain)
    yhat = model_full.predict(xtest)
    for i in index:
        if i in xtrain.index:
            xtrain_rm = xtrain.drop(index=i, axis=0)
            ytrain_rm = ytrain.drop(index=i, axis=0)
            model_rm = model.fit(xtrain_rm, ytrain_rm)
            yhat_rm = model_rm.predict(xtest)
            rgp_list.append((_rga(yhat, yhat_rm)))
        else:
            raise ValueError(f"index {i} is not available")
    return pd.DataFrame(rgp_list, index=index, columns=["RGP"])


def rgp_statistic_test(yhat, yhat_rm):
        """
        RGP based test for comparing the ordering of the ranks related to the full model with that of the
        reduced model without the observation of interest
        """
        yhat = pd.DataFrame(yhat).reset_index(drop=True)
        yhat_rm = pd.DataFrame(yhat_rm).reset_index(drop=True)
        jk_mat = pd.concat([yhat, yhat_rm], axis=1, keys=["yhat", "yhat_rm"])
        n = len(jk_mat)
        index = np.arange(n)
        jk_results = []
        for i in range(n):
            jk_sample = jk_mat.iloc[[x for x in index if x != i],:]
            jk_sample.reset_index(drop=True, inplace=True)
            jk_statistic = _delta_function(jk_sample)
            jk_results.append(jk_statistic)
        se = np.sqrt(((n-1)/n)*(sum([(x-np.mean(jk_results))**2 for x in jk_results])))
        z = (_den(yhat)-_num(yhat, yhat_rm))/se
        p_value = 2*scipy.stats.norm.cdf(-abs(z))
        return p_value


import pandas as pd
import numpy as np
import scipy
from .utils.util import _rga_delta_function, _rga


def rgr_single(xtrain, xtest, ytrain, ytest, model, variable, perturbation_percentage= 0.05):
    """
     ### RANK GRADUATION Robustness (RGR) MEASURE ###
     Function for the RGR measure computation regarding perturbation of a single variable
     Inputs: xtrain -> A dataframe including xtrain data used for traing the model; 
             xtest -> A dataframe including xtest data used for testing the model;
             ytrain -> Actual values of the target variable for traing the model; 
             ytest -> Actual values of the target variable for testing the model;
             model -> Model could be a regressor such as MLPRegressor() from sklearn.neural_network
             or a classifier for example RandomForestClassifier() from sklearn.ensemble;
             variable -> A string representing name of the column of interest for which we want to find the RGR value;
             perturbation_percentage -> The percentage indicating the percentile of interest for perturbation, by 
             default equal to 0.05. Therefore, the 5th and 95th percentiles are considered for the perturbation if a 
             specific percentage is not defined. The maximum value allowed to set for this input variable is 0.5.
    """ 

    if variable not in xtrain.columns:
        raise ValueError("The selected variable is not available.")
    if perturbation_percentage > 0.5 or perturbation_percentage < 0:
        raise ValueError("The perturbation percentage should be between 0 and 0.5.")
    xtrain = pd.DataFrame(xtrain).reset_index(drop=True)
    xtest = pd.DataFrame(xtest).reset_index(drop=True)
    ytrain = pd.DataFrame(ytrain).reset_index(drop=True)
    ytest = pd.DataFrame(ytest).reset_index(drop=True)
    model_full = model.fit(xtrain, ytrain)
    yhat = model_full.predict(xtest)
    perturbed_xtrain = perturb(xtrain, variable, perturbation_percentage)
    rf_perturbed = model.fit(perturbed_xtrain, ytrain)
    yhat_pert = rf_perturbed.predict(xtest)
    rgr_val = _rga(yhat, yhat_pert)
    return rgr_val


def rgr_all(xtrain, xtest, ytrain, ytest, model, perturbation_percentage= 0.05):
    """
     ### RANK GRADUATION Robustness (RGR) MEASURE ###
     Function for the RGR measure computation regarding perturbation of all variables
     Inputs: xtrain -> A dataframe including xtrain data used for traing the model; 
             xtest -> A dataframe including xtest data used for testing the model;
             ytrain -> Actual values of the target variable for traing the model; 
             ytest -> Actual values of the target variable for testing the model;
             model -> Model could be a regressor such as MLPRegressor() from sklearn.neural_network
             or a classifier for example RandomForestClassifier() from sklearn.ensemble;
             perturbation_percentage -> The percentage indicating the percentile of interest for perturbation, by 
             default equal to 0.05. Therefore, the 5th and 95th percentiles are considered for the perturbation if a 
             specific percentage is not defined. The maximum value allowed to set for this input variable is 0.5.
    """ 
    if perturbation_percentage > 0.5 or perturbation_percentage < 0:
        raise ValueError("The perturbation percentage should be between 0 and 0.5.")
    xtrain = pd.DataFrame(xtrain).reset_index(drop=True)
    xtest = pd.DataFrame(xtest).reset_index(drop=True)
    ytrain = pd.DataFrame(ytrain).reset_index(drop=True)
    ytest = pd.DataFrame(ytest).reset_index(drop=True)
    model_full = model.fit(xtrain, ytrain)
    yhat = model_full.predict(xtest)
    perturbed_xtrain = xtrain.copy()
    for i in xtrain.columns:
        perturbed_xtrain[i] = perturb(xtrain, i, perturbation_percentage)[i]
    rf_perturbed = model.fit(perturbed_xtrain, ytrain)
    yhat_pert = rf_perturbed.predict(xtest)
    rgr_val= _rga(yhat, yhat_pert)
    return rgr_val


def perturb(data, variable, perturbation_percentage= 0.05):
    """
     Function to perturb a single variable based on the replacement of the two percentiles selected using the 
     given perturbation_percentage.
     Inputs: data -> A dataframe including xtrain data used for traing the model; 
             variable -> A dataframe including xtest data used for testing the model;
             perturbation_percentage -> The percentage indicating the percentile of interest for perturbation, by 
             default equal to 0.05. Therefore, the 5th and 95th percentiles are considered for the perturbation if a 
             specific percentage is not defined. The maximum value allowed to set for this input variable is 0.5.
    """ 
    if perturbation_percentage > 0.5 or perturbation_percentage < 0:
        raise ValueError("The perturbation percentage should be between 0 and 0.5.")
    data = data.reset_index(drop=True)
    perturbed_variable = data.loc[:,variable]
    vals = [[i, values] for i, values in enumerate(perturbed_variable)]
    indices = [x[0] for x in sorted(vals, key= lambda item: item[1])]
    sorted_variable = [x[1] for x in sorted(vals, key= lambda item: item[1])]
    percentile_5_index = int(np.ceil(perturbation_percentage * len(sorted_variable)))
    percentile_95_index = int(np.ceil((1-perturbation_percentage) * len(sorted_variable)))
    values_before_5th_percentile = sorted_variable[:percentile_5_index]
    values_after_95th_percentile = sorted_variable[percentile_95_index:]
    n = min([len(values_before_5th_percentile), len(values_after_95th_percentile)])
    lowertail_indices = indices[0:n]
    uppertail_indices = (indices[-n:])
    uppertail_indices = uppertail_indices[::-1]
    new_variable = perturbed_variable.copy()
    for j in range(n):
        new_variable[lowertail_indices[j]] = perturbed_variable[uppertail_indices[j]]
        new_variable[uppertail_indices[j]] = perturbed_variable[lowertail_indices[j]]
    data.loc[:,variable] = new_variable
    return data


def rgr_statistic_test(yhat_mod1,yhat_mod2,yhat_pert_mod1,yhat_pert_mod2):
        """
        RGR based test for comparing the robustness of a model with that of a further compared model
        """
        yhat_mod1 = pd.DataFrame(yhat_mod1).reset_index(drop=True)
        yhat_mod2 = pd.DataFrame(yhat_mod2).reset_index(drop=True)
        yhat_pert_mod1 = pd.DataFrame(yhat_pert_mod1).reset_index(drop=True)
        yhat_pert_mod2 = pd.DataFrame(yhat_pert_mod2).reset_index(drop=True)
        jk_mat = pd.concat([yhat_mod1,yhat_mod2,yhat_pert_mod1,yhat_pert_mod2], axis=1, keys=["yhat_mod1", 
                                                                                        "yhat_mode2", 
                                                                                        "yhat_pert_mode1",
                                                                                        "yhat_pert_mode2"])
        n = len(jk_mat)
        index = np.arange(n)
        jk_results = []
        for i in range(n):
            jk_sample = jk_mat.iloc[[x for x in index if x != i],:]
            jk_sample.reset_index(drop=True, inplace=True)
            jk_statistic = _rga_delta_function(jk_sample, _rga)
            jk_results.append(jk_statistic)
        se = np.sqrt(((n-1)/n)*(sum([(x-np.mean(jk_results))**2 for x in jk_results])))
        z = (_rga(yhat_mod1, yhat_pert_mod1)- _rga(yhat_mod2, yhat_pert_mod2))/se
        p_value = 2*scipy.stats.norm.cdf(-abs(z))
        return p_value




import pandas as pd
import numpy as np
import scipy
from .utils.util import _rga_delta_function

def rga(y, yhat):
    """
     ### RANK GRADUATION ACCURACY (RGA) MEASURE ###
     Function for the RGE measure computation
     Inputs: y -> Actual values of the target variable; 
             yhat -> Predictions generated by the applied selected model
    """ 
    y = pd.DataFrame(y).reset_index(drop=True)
    yhat = pd.DataFrame(yhat).reset_index(drop=True)
    df = pd.concat([y,yhat], axis=1)
    df.columns = ["y", "yhat"]
    ryhat = yhat.rank(method="min")
    df["ryhat"] = ryhat
    support = df.groupby('ryhat')['y'].mean().reset_index(name='support')
    rord = list(range(len(y)))
    for jj in range(len(rord)):
        for ii in range(len(support)):
                if df["ryhat"][jj]== support['ryhat'][ii]:
                    rord[jj] = support['support'][ii]
    vals = [[i, values] for i, values in enumerate(df["yhat"])]
    ranks = [x[0] for x in sorted(vals, key= lambda item: item[1])]
    ystar = [rord[i] for i in ranks]
    I = list(range(len(y)))
    conc = 2*sum([I[i]*ystar[i] for i in range(len(I))])
    dec= 2*sum([sorted(df["y"], reverse=True)[i]*I[i] for i in range(len(I))]) 
    inc = 2*sum([sorted(df["y"])[i]*I[i] for i in range(len(I))]) 
    RGA=(conc-dec)/(inc-dec)
    return RGA 

def rga_statistic_test(y, yhat_rm, yhat_cm):
        """
        RGA based test for comparing the predictive accuracy of a reduced model with that of a more complex model
        """
        y = pd.DataFrame(y).reset_index(drop=True)
        yhat_rm = pd.DataFrame(yhat_rm).reset_index(drop=True)
        yhat_cm = pd.DataFrame(yhat_cm).reset_index(drop=True)
        jk_mat = pd.concat([y,yhat_rm, yhat_cm], axis=1)
        jk_mat.columns = ["y", "yhat_rm", "yhat_cm"]
        n = len(jk_mat)
        index = np.arange(n)
        jk_results = []
        for i in range(n):
            jk_sample = jk_mat.drop(labels= i)
            jk_sample.reset_index(drop=True, inplace=True)
            jk_statistic = _rga_delta_function(jk_sample, rga)
            jk_results.append(jk_statistic)
        se = np.sqrt(((n-1)/n)*(sum([(x-np.mean(jk_results))**2 for x in jk_results])))
        z = (rga(y,yhat_cm)-rga(y,yhat_rm))/se
        p_value = 2*scipy.stats.norm.cdf(-abs(z))
        return p_value


from setuptools import setup, find_packages
from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent

# Open README.md with UTF-8 encoding
with open(this_directory / "README.md", encoding="utf-8") as f:
    long_description = f.read()
    
VERSION = '0.2.0'
DESCRIPTION = 'SAFE AI package to measure risks of AI models'

# Setting up
setup(
    name="safeaipackage",
    version=VERSION,
    author="Golnoosh Babaei",
    author_email="<golnoosh.babaei93@gmail.com>",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



"""This is a helper module with utility classes and functions."""
import pandas as pd
from sklearn.ensemble import RandomForestClassifier

def _rga_delta_function(data, func):
        result = (func(data.iloc[:,0], data.iloc[:,2]))-(func(data.iloc[:,0], data.iloc[:,1]))
        return result


def _rga(y, yhat):
    y = pd.DataFrame(y).reset_index(drop=True)
    yhat = pd.DataFrame(yhat).reset_index(drop=True)
    df = pd.concat([y,yhat], axis=1)
    df.columns = ["y", "yhat"]
    ryhat = yhat.rank(method="min")
    df["ryhat"] = ryhat
    support = df.groupby('ryhat')['y'].mean().reset_index(name='support')
    rord = list(range(len(y)))
    for jj in range(len(rord)):
        for ii in range(len(support)):
                if df["ryhat"][jj]== support['ryhat'][ii]:
                    rord[jj] = support['support'][ii]
    vals = [[i, values] for i, values in enumerate(df["yhat"])]
    ranks = [x[0] for x in sorted(vals, key= lambda item: item[1])]
    ystar = [rord[i] for i in ranks]
    I = list(range(len(y)))
    conc = 2*sum([I[i]*ystar[i] for i in range(len(I))])
    dec= 2*sum([sorted(df["y"], reverse=True)[i]*I[i] for i in range(len(I))]) 
    inc = 2*sum([sorted(df["y"])[i]*I[i] for i in range(len(I))]) 
    RGA=(conc-dec)/(inc-dec)
    return RGA


def _num(yhat, yhat_xk):
    yhat = pd.DataFrame(yhat).reset_index(drop=True)
    yhat_xk = pd.DataFrame(yhat_xk).reset_index(drop=True)
    df = pd.concat([yhat,yhat_xk], axis=1)
    df.columns = ["yhat", "yhat_xk"]
    ryhat_xk = yhat_xk.rank(method="min")
    df["ryhat_xk"] = ryhat_xk
    support = df.groupby('ryhat_xk')['yhat'].mean().reset_index(name='support')
    rord = list(range(len(yhat)))
    for jj in range(len(rord)):
        for ii in range(len(support)):
                if df["ryhat_xk"][jj]== support['ryhat_xk'][ii]:
                    rord[jj] = support['support'][ii]
    vals = [[i, values] for i, values in enumerate(df["yhat_xk"])]
    ranks = [x[0] for x in sorted(vals, key= lambda item: item[1])]
    ystar = [rord[i] for i in ranks]
    I = list(range(len(ystar)))
    conc = 2*sum([I[i]*ystar[i] for i in range(len(I))])
    dec= 2*sum([sorted(df["yhat"], reverse=True)[i]*I[i] for i in range(len(I))]) 
    inc = 2*sum([sorted(df["yhat"])[i]*I[i] for i in range(len(I))]) 
    RGE_num=(conc-dec)
    return RGE_num


def _den(yhat):
    yhat = pd.DataFrame(yhat)
    yhat.columns = ["yhat"]
    I = list(range(len(yhat)))
    dec= 2*sum([yhat.sort_values(by="yhat", ascending=False, ignore_index=True).iloc[i,0]*I[i] for i in range(len(I))]) 
    inc = 2*sum([yhat.sort_values(by="yhat", ignore_index=True).iloc[i,0]*I[i] for i in range(len(I))]) 
    RGE_den=(inc-dec)
    return RGE_den


def _delta_function(data):
    return _den(data.iloc[:,0])-_num(data.iloc[:,0], data.iloc[:,1])
